@@ -1,4 +1,3 @@
-// SPDX-License-Identifier: GPL-2.0-or-later
 /*
  *
  *  BlueZ - Bluetooth protocol stack for Linux
@@ -8,36 +7,42 @@
  *  Copyright (C) 2011  BMW Car IT GmbH. All rights reserved.
  *
  *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; either version 2 of the License, or
+ *  (at your option) any later version.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ *
  */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
 
-#define _GNU_SOURCE
 #include <errno.h>
 #include <inttypes.h>
 
 #include <glib.h>
+#include <gdbus/gdbus.h>
 
-#include "lib/bluetooth.h"
-#include "lib/sdp.h"
 #include "lib/uuid.h"
-
-#include "gdbus/gdbus.h"
-
 #include "src/plugin.h"
 #include "src/adapter.h"
 #include "src/device.h"
 #include "src/dbus-common.h"
 #include "src/profile.h"
 
-#include "src/uuid-helper.h"
-#include "src/log.h"
-#include "src/error.h"
-#include "src/shared/util.h"
-#include "src/shared/queue.h"
-
+#include "glib-helper.h"
+#include "log.h"
+#include "error.h"
 #include "avdtp.h"
 #include "media.h"
 #include "transport.h"
@@ -50,28 +55,14 @@
 
 #define REQUEST_TIMEOUT (3 * 1000)		/* 3 seconds */
 
-struct media_app {
-	struct media_adapter	*adapter;
-	GDBusClient		*client;
-	DBusMessage		*reg;
-	char			*sender;	/* Application bus id */
-	char			*path;		/* Application object path */
-	struct queue		*proxies;	/* Application proxies */
-	struct queue		*endpoints;	/* Application endpoints */
-	struct queue		*players;	/* Application players */
-	int			err;
-};
-
 struct media_adapter {
 	struct btd_adapter	*btd_adapter;
-	struct queue		*apps;		/* Application list */
 	GSList			*endpoints;	/* Endpoints list */
 	GSList			*players;	/* Players list */
 };
 
 struct endpoint_request {
 	struct media_endpoint	*endpoint;
-	struct media_transport	*transport;
 	DBusMessage		*msg;
 	DBusPendingCall		*call;
 	media_endpoint_cb_t	cb;
@@ -108,14 +99,13 @@ struct media_player {
 	char			*status;
 	uint32_t		position;
 	uint32_t		duration;
-	int8_t			volume;
+	uint8_t			volume;
 	GTimer			*timer;
 	bool			play;
 	bool			pause;
 	bool			next;
 	bool			previous;
 	bool			control;
-	char			*name;
 };
 
 static GSList *adapters = NULL;
@@ -196,9 +186,8 @@ static struct media_endpoint *media_adapter_find_endpoint(
 	return NULL;
 }
 
-static void media_endpoint_remove(void *data)
+static void media_endpoint_remove(struct media_endpoint *endpoint)
 {
-	struct media_endpoint *endpoint = data;
 	struct media_adapter *adapter = endpoint->adapter;
 
 	if (endpoint->sep) {
@@ -227,20 +216,6 @@ static void media_endpoint_exit(DBusConnection *connection, void *user_data)
 	media_endpoint_remove(endpoint);
 }
 
-static struct media_adapter *find_adapter(struct btd_device *device)
-{
-	GSList *l;
-
-	for (l = adapters; l; l = l->next) {
-		struct media_adapter *adapter = l->data;
-
-		if (adapter->btd_adapter == device_get_adapter(device))
-			return adapter;
-	}
-
-	return NULL;
-}
-
 static void clear_configuration(struct media_endpoint *endpoint,
 					struct media_transport *transport)
 {
@@ -293,21 +268,15 @@ static void endpoint_reply(DBusPendingCall *call, void *user_data)
 
 		/* Clear endpoint configuration in case of NO_REPLY error */
 		if (dbus_error_has_name(&err, DBUS_ERROR_NO_REPLY)) {
+			if (request->cb)
+				request->cb(endpoint, NULL, size,
+							request->user_data);
 			clear_endpoint(endpoint);
 			dbus_message_unref(reply);
 			dbus_error_free(&err);
 			return;
 		}
 
-		if (dbus_message_is_method_call(request->msg,
-					MEDIA_ENDPOINT_INTERFACE,
-					"SetConfiguration")) {
-			if (request->transport == NULL)
-				error("Expected to destroy transport");
-			else
-				media_transport_destroy(request->transport);
-		}
-
 		dbus_error_free(&err);
 		goto done;
 	}
@@ -347,7 +316,6 @@ done:
 
 static gboolean media_endpoint_async_call(DBusMessage *msg,
 					struct media_endpoint *endpoint,
-					struct media_transport *transport,
 					media_endpoint_cb_t cb,
 					void *user_data,
 					GDestroyNotify destroy)
@@ -369,7 +337,6 @@ static gboolean media_endpoint_async_call(DBusMessage *msg,
 									NULL);
 
 	request->endpoint = endpoint;
-	request->transport = transport;
 	request->msg = msg;
 	request->cb = cb;
 	request->destroy = destroy;
@@ -405,8 +372,7 @@ static gboolean select_configuration(struct media_endpoint *endpoint,
 					&capabilities, length,
 					DBUS_TYPE_INVALID);
 
-	return media_endpoint_async_call(msg, endpoint, NULL,
-						cb, user_data, destroy);
+	return media_endpoint_async_call(msg, endpoint, cb, user_data, destroy);
 }
 
 static int transport_device_cmp(gconstpointer data, gconstpointer user_data)
@@ -440,33 +406,6 @@ struct a2dp_config_data {
 	a2dp_endpoint_config_t cb;
 };
 
-int8_t media_player_get_device_volume(struct btd_device *device)
-{
-	struct avrcp_player *target_player;
-	struct media_adapter *adapter;
-	GSList *l;
-
-	if (!device)
-		return -1;
-
-	target_player = avrcp_get_target_player_by_device(device);
-	if (!target_player)
-		return -1;
-
-	adapter = find_adapter(device);
-	if (!adapter)
-		return -1;
-
-	for (l = adapter->players; l; l = l->next) {
-		struct media_player *mp = l->data;
-
-		if (mp->player == target_player)
-			return mp->volume;
-	}
-
-	return -1;
-}
-
 static gboolean set_configuration(struct media_endpoint *endpoint,
 					uint8_t *configuration, size_t size,
 					media_endpoint_cb_t cb,
@@ -480,22 +419,17 @@ static gboolean set_configuration(struct media_endpoint *endpoint,
 	const char *path;
 	DBusMessageIter iter;
 	struct media_transport *transport;
-	int8_t init_volume;
 
 	transport = find_device_transport(endpoint, device);
 
 	if (transport != NULL)
 		return FALSE;
 
-	transport = media_transport_create(device,
-					a2dp_setup_remote_path(data->setup),
-					configuration, size, endpoint);
+	transport = media_transport_create(device, configuration, size,
+								endpoint);
 	if (transport == NULL)
 		return FALSE;
 
-	init_volume = media_player_get_device_volume(device);
-	media_transport_update_volume(transport, init_volume);
-
 	msg = dbus_message_new_method_call(endpoint->sender, endpoint->path,
 						MEDIA_ENDPOINT_INTERFACE,
 						"SetConfiguration");
@@ -514,8 +448,7 @@ static gboolean set_configuration(struct media_endpoint *endpoint,
 
 	g_dbus_get_properties(conn, path, "org.bluez.MediaTransport1", &iter);
 
-	return media_endpoint_async_call(msg, endpoint, transport,
-						cb, user_data, destroy);
+	return media_endpoint_async_call(msg, endpoint, cb, user_data, destroy);
 }
 
 static void release_endpoint(struct media_endpoint *endpoint)
@@ -551,13 +484,6 @@ static const char *get_name(struct a2dp_sep *sep, void *user_data)
 	return endpoint->sender;
 }
 
-static const char *get_path(struct a2dp_sep *sep, void *user_data)
-{
-	struct media_endpoint *endpoint = user_data;
-
-	return endpoint->path;
-}
-
 static size_t get_capabilities(struct a2dp_sep *sep, uint8_t **capabilities,
 							void *user_data)
 {
@@ -603,9 +529,8 @@ static void config_cb(struct media_endpoint *endpoint, void *ret, int size,
 							void *user_data)
 {
 	struct a2dp_config_data *data = user_data;
-	gboolean *ret_value = ret;
 
-	data->cb(data->setup, ret_value ? *ret_value : FALSE);
+	data->cb(data->setup, ret ? TRUE : FALSE);
 }
 
 static int set_config(struct a2dp_sep *sep, uint8_t *configuration,
@@ -648,7 +573,6 @@ static void set_delay(struct a2dp_sep *sep, uint16_t delay, void *user_data)
 
 static struct a2dp_endpoint a2dp_endpoint = {
 	.get_name = get_name,
-	.get_path = get_path,
 	.get_capabilities = get_capabilities,
 	.select_configuration = select_config,
 	.set_configuration = set_config,
@@ -692,6 +616,20 @@ static gboolean endpoint_init_a2dp_sink(struct media_endpoint *endpoint,
 	return TRUE;
 }
 
+static struct media_adapter *find_adapter(struct btd_device *device)
+{
+	GSList *l;
+
+	for (l = adapters; l; l = l->next) {
+		struct media_adapter *adapter = l->data;
+
+		if (adapter->btd_adapter == device_get_adapter(device))
+			return adapter;
+	}
+
+	return NULL;
+}
+
 static bool endpoint_properties_exists(const char *uuid,
 						struct btd_device *dev,
 						void *user_data)
@@ -1022,7 +960,6 @@ static void media_player_free(gpointer data)
 	g_free(mp->sender);
 	g_free(mp->path);
 	g_free(mp->status);
-	g_free(mp->name);
 	g_free(mp);
 }
 
@@ -1043,10 +980,8 @@ static void media_player_destroy(struct media_player *mp)
 	media_player_free(mp);
 }
 
-static void media_player_remove(void *data)
+static void media_player_remove(struct media_player *mp)
 {
-	struct media_player *mp = data;
-
 	info("Player unregistered: sender=%s path=%s", mp->sender, mp->path);
 
 	media_player_destroy(mp);
@@ -1073,16 +1008,6 @@ static const char *get_setting(const char *key, void *user_data)
 	return g_hash_table_lookup(mp->settings, key);
 }
 
-static const char *get_player_name(void *user_data)
-{
-	struct media_player *mp = user_data;
-
-	if (!mp->name)
-		return "Player";
-
-	return mp->name;
-}
-
 static void set_shuffle_setting(DBusMessageIter *iter, const char *value)
 {
 	const char *key = "Shuffle";
@@ -1106,8 +1031,6 @@ static const char *repeat_to_loop_status(const char *value)
 		return "Track";
 	else if (strcasecmp(value, "alltracks") == 0)
 		return "Playlist";
-	else if (strcasecmp(value, "group") == 0)
-		return "Playlist";
 
 	return NULL;
 }
@@ -1133,17 +1056,12 @@ static int set_setting(const char *key, const char *value, void *user_data)
 	const char *iface = MEDIA_PLAYER_INTERFACE;
 	DBusMessage *msg;
 	DBusMessageIter iter;
-	const char *curval;
 
 	DBG("%s = %s", key, value);
 
-	curval = g_hash_table_lookup(mp->settings, key);
-	if (!curval)
+	if (!g_hash_table_lookup(mp->settings, key))
 		return -EINVAL;
 
-	if (strcasecmp(curval, value) == 0)
-		return 0;
-
 	msg = dbus_message_new_method_call(mp->sender, mp->path,
 					DBUS_INTERFACE_PROPERTIES, "Set");
 	if (msg == NULL) {
@@ -1231,14 +1149,30 @@ static uint32_t get_duration(void *user_data)
 	return mp->duration;
 }
 
-static void set_volume(int8_t volume, struct btd_device *dev, void *user_data)
+static void set_volume(uint8_t volume, struct btd_device *dev, void *user_data)
 {
 	struct media_player *mp = user_data;
+	GSList *l;
 
 	if (mp->volume == volume)
 		return;
 
 	mp->volume = volume;
+
+	for (l = mp->adapter->endpoints; l; l = l->next) {
+		struct media_endpoint *endpoint = l->data;
+		struct media_transport *transport;
+
+		/* Volume is A2DP only */
+		if (endpoint->sep == NULL)
+			continue;
+
+		transport = find_device_transport(endpoint, dev);
+		if (transport == NULL)
+			continue;
+
+		media_transport_update_volume(transport, volume);
+	}
 }
 
 static bool media_player_send(struct media_player *mp, const char *name)
@@ -1281,7 +1215,7 @@ static bool stop(void *user_data)
 	return media_player_send(mp, "Stop");
 }
 
-static bool apause(void *user_data)
+static bool pause(void *user_data)
 {
 	struct media_player *mp = user_data;
 
@@ -1327,11 +1261,10 @@ static struct avrcp_player_cb player_cb = {
 	.get_position = get_position,
 	.get_duration = get_duration,
 	.get_status = get_status,
-	.get_name = get_player_name,
 	.set_volume = set_volume,
 	.play = play,
 	.stop = stop,
-	.pause = apause,
+	.pause = pause,
 	.next = next,
 	.previous = previous,
 };
@@ -1374,7 +1307,7 @@ static gboolean set_position(struct media_player *mp, DBusMessageIter *iter)
 	const char *status;
 
 	if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_INT64)
-		return FALSE;
+			return FALSE;
 
 	dbus_message_iter_get_basic(iter, &value);
 
@@ -1663,25 +1596,6 @@ static gboolean set_flag(struct media_player *mp, DBusMessageIter *iter,
 	return TRUE;
 }
 
-static gboolean set_name(struct media_player *mp, DBusMessageIter *iter)
-{
-	const char *value;
-
-	if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_STRING)
-		return FALSE;
-
-	dbus_message_iter_get_basic(iter, &value);
-
-	if (g_strcmp0(mp->name, value) == 0)
-		return TRUE;
-
-	g_free(mp->name);
-
-	mp->name = g_strdup(value);
-
-	return TRUE;
-}
-
 static gboolean set_player_property(struct media_player *mp, const char *key,
 							DBusMessageIter *entry)
 {
@@ -1722,9 +1636,6 @@ static gboolean set_player_property(struct media_player *mp, const char *key,
 	if (strcasecmp(key, "CanControl") == 0)
 		return set_flag(mp, &var, &mp->control);
 
-	if (strcasecmp(key, "Identity") == 0)
-		return set_name(mp, &var);
-
 	DBG("%s not supported, ignoring", key);
 
 	return TRUE;
@@ -1811,7 +1722,6 @@ static struct media_player *media_player_create(struct media_adapter *adapter,
 	mp->sender = g_strdup(sender);
 	mp->path = g_strdup(path);
 	mp->timer = g_timer_new();
-	mp->volume = -1;
 
 	mp->watch = g_dbus_add_disconnect_watch(conn, sender,
 						media_player_exit, mp,
@@ -1904,479 +1814,17 @@ static DBusMessage *unregister_player(DBusConnection *conn, DBusMessage *msg,
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
-static void app_free(void *data)
-{
-	struct media_app *app = data;
-
-	queue_destroy(app->endpoints, media_endpoint_remove);
-	queue_destroy(app->players, media_player_remove);
-
-	if (app->client) {
-		g_dbus_client_set_disconnect_watch(app->client, NULL, NULL);
-		g_dbus_client_set_proxy_handlers(app->client, NULL, NULL,
-								NULL, NULL);
-		g_dbus_client_set_ready_watch(app->client, NULL, NULL);
-		g_dbus_client_unref(app->client);
-	}
-
-	if (app->reg)
-		dbus_message_unref(app->reg);
-
-	g_free(app->sender);
-	g_free(app->path);
-
-	free(app);
-}
-
-static void client_disconnect_cb(DBusConnection *conn, void *user_data)
-{
-	struct media_app *app = user_data;
-	struct media_adapter *adapter = app->adapter;
-
-	DBG("Client disconnected");
-
-	if (queue_remove(adapter->apps, app))
-		app_free(app);
-}
-
-static void app_register_endpoint(void *data, void *user_data)
-{
-	struct media_app *app = user_data;
-	GDBusProxy *proxy = data;
-	const char *iface = g_dbus_proxy_get_interface(proxy);
-	const char *path = g_dbus_proxy_get_path(proxy);
-	const char *uuid;
-	gboolean delay_reporting = FALSE;
-	uint8_t codec;
-	uint8_t *capabilities = NULL;
-	int size = 0;
-	DBusMessageIter iter, array;
-	struct media_endpoint *endpoint;
-
-	if (app->err)
-		return;
-
-	if (strcmp(iface, MEDIA_ENDPOINT_INTERFACE))
-		return;
-
-	/* Parse properties */
-	if (!g_dbus_proxy_get_property(proxy, "UUID", &iter))
-		goto fail;
-
-	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
-		goto fail;
-
-	dbus_message_iter_get_basic(&iter, &uuid);
-
-	if (!g_dbus_proxy_get_property(proxy, "Codec", &iter))
-		goto fail;
-
-	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_BYTE)
-		goto fail;
-
-	dbus_message_iter_get_basic(&iter, &codec);
-
-	/* DelayReporting and Capabilities are considered optional */
-	if (g_dbus_proxy_get_property(proxy, "DelayReporting", &iter))	{
-		if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_BOOLEAN)
-			goto fail;
-
-		dbus_message_iter_get_basic(&iter, &delay_reporting);
-	}
-
-	if (g_dbus_proxy_get_property(proxy, "Capabilities", &iter)) {
-		if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_ARRAY)
-			goto fail;
-
-		dbus_message_iter_recurse(&iter, &array);
-		dbus_message_iter_get_fixed_array(&array, &capabilities, &size);
-	}
-
-	endpoint = media_endpoint_create(app->adapter, app->sender, path, uuid,
-					delay_reporting, codec, capabilities,
-					size, &app->err);
-	if (!endpoint) {
-		error("Unable to register endpoint %s:%s: %s", app->sender,
-						path, strerror(-app->err));
-		return;
-	}
-
-	queue_push_tail(app->endpoints, endpoint);
-
-	return;
-
-fail:
-	app->err = -EINVAL;
-}
-
-static void app_register_player(void *data, void *user_data)
-{
-	struct media_app *app = user_data;
-	GDBusProxy *proxy = data;
-	const char *iface = g_dbus_proxy_get_interface(proxy);
-	const char *path = g_dbus_proxy_get_path(proxy);
-	struct media_player *player;
-	DBusMessageIter iter;
-
-	if (app->err)
-		return;
-
-	if (strcmp(iface, MEDIA_PLAYER_INTERFACE))
-		return;
-
-	player = media_player_create(app->adapter, app->sender, path,
-							&app->err);
-	if (!player)
-		return;
-
-	if (g_dbus_proxy_get_property(proxy, "PlaybackStatus", &iter)) {
-		if (!set_status(player, &iter))
-			goto fail;
-	}
-
-	if (g_dbus_proxy_get_property(proxy, "Position", &iter)) {
-		if (!set_position(player, &iter))
-			goto fail;
-	}
-
-	if (g_dbus_proxy_get_property(proxy, "Metadata", &iter)) {
-		if (!parse_player_metadata(player, &iter))
-			goto fail;
-	}
-
-	if (g_dbus_proxy_get_property(proxy, "Shuffle", &iter)) {
-		if (!set_shuffle(player, &iter))
-			goto fail;
-	}
-
-	if (g_dbus_proxy_get_property(proxy, "LoopStatus", &iter)) {
-		if (!set_repeat(player, &iter))
-			goto fail;
-	}
-
-	if (g_dbus_proxy_get_property(proxy, "CanPlay", &iter)) {
-		if (!set_flag(player, &iter, &player->play))
-			goto fail;
-	}
-
-	if (g_dbus_proxy_get_property(proxy, "CanPause", &iter)) {
-		if (!set_flag(player, &iter, &player->pause))
-			goto fail;
-	}
-
-	if (g_dbus_proxy_get_property(proxy, "CanGoNext", &iter)) {
-		if (!set_flag(player, &iter, &player->next))
-			goto fail;
-	}
-
-	if (g_dbus_proxy_get_property(proxy, "CanGoPrevious", &iter)) {
-		if (!set_flag(player, &iter, &player->previous))
-			goto fail;
-	}
-
-	if (g_dbus_proxy_get_property(proxy, "CanControl", &iter)) {
-		if (!set_flag(player, &iter, &player->control))
-			goto fail;
-	}
-
-	if (g_dbus_proxy_get_property(proxy, "Identity", &iter)) {
-		if (!set_name(player, &iter))
-			goto fail;
-	}
-
-	queue_push_tail(app->players, player);
-
-	return;
-fail:
-	app->err = -EINVAL;
-	error("Unable to register player %s:%s: %s", app->sender, path,
-							strerror(-app->err));
-	media_player_destroy(player);
-}
-
-static void remove_app(void *data)
-{
-	struct media_app *app = data;
-
-	/*
-	 * Set callback to NULL to avoid potential race condition
-	 * when calling remove_app and GDBusClient unref.
-	 */
-	g_dbus_client_set_disconnect_watch(app->client, NULL, NULL);
-
-	/*
-	 * Set proxy handlers to NULL, so that this gets called only once when
-	 * the first proxy that belongs to this service gets removed.
-	 */
-	g_dbus_client_set_proxy_handlers(app->client, NULL, NULL, NULL, NULL);
-
-
-	queue_remove(app->adapter->apps, app);
-
-	app_free(app);
-}
-
-static void client_ready_cb(GDBusClient *client, void *user_data)
-{
-	struct media_app *app = user_data;
-	DBusMessage *reply;
-	bool fail = false;
-
-	/*
-	 * Process received objects
-	 */
-	if (queue_isempty(app->proxies)) {
-		error("No object received");
-		fail = true;
-		reply = btd_error_failed(app->reg, "No object received");
-		goto reply;
-	}
-
-	queue_foreach(app->proxies, app_register_endpoint, app);
-	queue_foreach(app->proxies, app_register_player, app);
-
-	if (app->err) {
-		if (app->err == -EPROTONOSUPPORT)
-			reply = btd_error_not_supported(app->reg);
-		else
-			reply = btd_error_invalid_args(app->reg);
-		goto reply;
-	}
-
-	if ((queue_isempty(app->endpoints) && queue_isempty(app->players))) {
-		error("No valid external Media objects found");
-		fail = true;
-		reply = btd_error_failed(app->reg,
-					"No valid media object found");
-		goto reply;
-	}
-
-	DBG("Media application registered: %s:%s", app->sender, app->path);
-
-	reply = dbus_message_new_method_return(app->reg);
-
-reply:
-	g_dbus_send_message(btd_get_dbus_connection(), reply);
-	dbus_message_unref(app->reg);
-	app->reg = NULL;
-
-	if (fail)
-		remove_app(app);
-}
-
-static void proxy_added_cb(GDBusProxy *proxy, void *user_data)
-{
-	struct media_app *app = user_data;
-	const char *iface, *path;
-
-	if (app->err)
-		return;
-
-	queue_push_tail(app->proxies, proxy);
-
-	iface = g_dbus_proxy_get_interface(proxy);
-	path = g_dbus_proxy_get_path(proxy);
-
-	DBG("Proxy added: %s, iface: %s", path, iface);
-}
-
-static bool match_endpoint_by_path(const void *a, const void *b)
-{
-	const struct media_endpoint *endpoint = a;
-	const char *path = b;
-
-	return !strcmp(endpoint->path, path);
-}
-
-static bool match_player_by_path(const void *a, const void *b)
-{
-	const struct media_player *player = a;
-	const char *path = b;
-
-	return !strcmp(player->path, path);
-}
-
-static void proxy_removed_cb(GDBusProxy *proxy, void *user_data)
-{
-	struct media_app *app = user_data;
-	struct media_endpoint *endpoint;
-	struct media_player *player;
-	const char *iface, *path;
-
-	iface = g_dbus_proxy_get_interface(proxy);
-	path = g_dbus_proxy_get_path(proxy);
-
-	if (!strcmp(iface, MEDIA_ENDPOINT_INTERFACE)) {
-		endpoint = queue_remove_if(app->endpoints,
-						match_endpoint_by_path,
-						(void *) path);
-		if (!endpoint)
-			return;
-
-		if (!g_slist_find(app->adapter->endpoints, endpoint))
-			return;
-
-		DBG("Proxy removed - removing endpoint: %s", endpoint->path);
-
-		media_endpoint_remove(endpoint);
-	} else if (!strcmp(iface, MEDIA_PLAYER_INTERFACE)) {
-		player = queue_remove_if(app->players, match_player_by_path,
-						(void *) path);
-		if (!player)
-			return;
-
-		if (!g_slist_find(app->adapter->players, player))
-			return;
-
-		DBG("Proxy removed - removing player: %s", player->path);
-
-		media_player_remove(player);
-	}
-}
-
-static struct media_app *create_app(DBusConnection *conn, DBusMessage *msg,
-							const char *path)
-{
-	struct media_app *app;
-	const char *sender = dbus_message_get_sender(msg);
-
-	if (!path || !g_str_has_prefix(path, "/"))
-		return NULL;
-
-	app = new0(struct media_app, 1);
-
-	app->client = g_dbus_client_new_full(conn, sender, path, path);
-	if (!app->client)
-		goto fail;
-
-	app->sender = g_strdup(sender);
-	if (!app->sender)
-		goto fail;
-
-	app->path = g_strdup(path);
-	if (!app->path)
-		goto fail;
-
-	app->proxies = queue_new();
-	app->endpoints = queue_new();
-	app->players = queue_new();
-	app->reg = dbus_message_ref(msg);
-
-	g_dbus_client_set_disconnect_watch(app->client, client_disconnect_cb,
-									app);
-	g_dbus_client_set_proxy_handlers(app->client, proxy_added_cb,
-					proxy_removed_cb, NULL, app);
-	g_dbus_client_set_ready_watch(app->client, client_ready_cb, app);
-
-	return app;
-
-fail:
-	app_free(app);
-	return NULL;
-}
-
-struct match_data {
-	const char *path;
-	const char *sender;
-};
-
-static bool match_app(const void *a, const void *b)
-{
-	const struct media_app *app = a;
-	const struct match_data *data = b;
-
-	return g_strcmp0(app->path, data->path) == 0 &&
-				g_strcmp0(app->sender, data->sender) == 0;
-}
-
-static DBusMessage *register_app(DBusConnection *conn, DBusMessage *msg,
-							void *user_data)
-{
-	struct media_adapter *adapter = user_data;
-	const char *sender = dbus_message_get_sender(msg);
-	DBusMessageIter args;
-	const char *path;
-	struct media_app *app;
-	struct match_data match_data;
-
-	if (!dbus_message_iter_init(msg, &args))
-		return btd_error_invalid_args(msg);
-
-	if (dbus_message_iter_get_arg_type(&args) != DBUS_TYPE_OBJECT_PATH)
-		return btd_error_invalid_args(msg);
-
-	dbus_message_iter_get_basic(&args, &path);
-
-	match_data.path = path;
-	match_data.sender = sender;
-
-	if (queue_find(adapter->apps, match_app, &match_data))
-		return btd_error_already_exists(msg);
-
-	dbus_message_iter_next(&args);
-	if (dbus_message_iter_get_arg_type(&args) != DBUS_TYPE_ARRAY)
-		return btd_error_invalid_args(msg);
-
-	app = create_app(conn, msg, path);
-	if (!app)
-		return btd_error_failed(msg, "Failed to register application");
-
-	DBG("Registering application: %s:%s", sender, path);
-
-	app->adapter = adapter;
-	queue_push_tail(adapter->apps, app);
-
-	return NULL;
-}
-
-static DBusMessage *unregister_app(DBusConnection *conn, DBusMessage *msg,
-							void *user_data)
-{
-	struct media_adapter *adapter = user_data;
-	const char *sender = dbus_message_get_sender(msg);
-	const char *path;
-	DBusMessageIter args;
-	struct media_app *app;
-	struct match_data match_data;
-
-	if (!dbus_message_iter_init(msg, &args))
-		return btd_error_invalid_args(msg);
-
-	if (dbus_message_iter_get_arg_type(&args) != DBUS_TYPE_OBJECT_PATH)
-		return btd_error_invalid_args(msg);
-
-	dbus_message_iter_get_basic(&args, &path);
-
-	match_data.path = path;
-	match_data.sender = sender;
-
-	app = queue_remove_if(adapter->apps, match_app, &match_data);
-	if (!app)
-		return btd_error_does_not_exist(msg);
-
-	app_free(app);
-
-	return dbus_message_new_method_return(msg);
-}
-
 static const GDBusMethodTable media_methods[] = {
 	{ GDBUS_METHOD("RegisterEndpoint",
 		GDBUS_ARGS({ "endpoint", "o" }, { "properties", "a{sv}" }),
 		NULL, register_endpoint) },
 	{ GDBUS_METHOD("UnregisterEndpoint",
 		GDBUS_ARGS({ "endpoint", "o" }), NULL, unregister_endpoint) },
-	{ GDBUS_METHOD("RegisterPlayer",
+	{ GDBUS_EXPERIMENTAL_METHOD("RegisterPlayer",
 		GDBUS_ARGS({ "player", "o" }, { "properties", "a{sv}" }),
 		NULL, register_player) },
-	{ GDBUS_METHOD("UnregisterPlayer",
+	{ GDBUS_EXPERIMENTAL_METHOD("UnregisterPlayer",
 		GDBUS_ARGS({ "player", "o" }), NULL, unregister_player) },
-	{ GDBUS_ASYNC_METHOD("RegisterApplication",
-					GDBUS_ARGS({ "application", "o" },
-						{ "options", "a{sv}" }),
-					NULL, register_app) },
-	{ GDBUS_ASYNC_METHOD("UnregisterApplication",
-					GDBUS_ARGS({ "application", "o" }),
-					NULL, unregister_app) },
 	{ },
 };
 
@@ -2384,8 +1832,6 @@ static void path_free(void *data)
 {
 	struct media_adapter *adapter = data;
 
-	queue_destroy(adapter->apps, app_free);
-
 	while (adapter->endpoints)
 		release_endpoint(adapter->endpoints->data);
 
@@ -2404,7 +1850,6 @@ int media_register(struct btd_adapter *btd_adapter)
 
 	adapter = g_new0(struct media_adapter, 1);
 	adapter->btd_adapter = btd_adapter_ref(btd_adapter);
-	adapter->apps = queue_new();
 
 	if (!g_dbus_register_interface(btd_get_dbus_connection(),
 					adapter_get_path(btd_adapter),
